package se.chalmers.cs.gf.tramdemo;

import java.awt.*;
import java.awt.event.*;
import java.awt.geom.AffineTransform;
import java.awt.geom.NoninvertibleTransformException;
import javax.swing.*;
import java.util.List;
import java.util.LinkedList;

/**
 *  All the instance methods in this class must be called from the 
 *  Swing event dispatching thread.
 */
public class MapPanel extends JPanel {

        private static final int NODE_RADIUS = 5;
        private static final int LABEL_X_OFF = 5;
        private static final int LABEL_Y_OFF = -5;

        private List<Node> nodes;

        private List<Edge> edges;

        private List<Edge> path;

        /** Margins in panel coordinates. */
        private int marginLeft, marginRight, marginTop, marginBottom;

        private Dimension mapSize;
        
        private double threshold;

        private AffineTransform mapToScreenTransform;

        private AffineTransform screenToMapTransform;

        private Stroke edgeStroke;
        private Stroke pathStroke;

        private Font nodeLabelFont;
        private Font pathLabelFont;

        private ImageIcon backgroundImage;

        /**
         *  @param mapSize The size of the map (in map coordinate space)
         *  @param panelSize The size of the panel (in graphics user coordinate space)
         *  @param threshold The size of the hotspots around nodes. Just used for
         *  presentation, actual click to node resolution is done by CoordMap
         */
        public MapPanel(Dimension mapSize, Dimension panelSize, double threshold, ImageIcon backgroundImage) {
                this.nodes = new LinkedList<Node>();
                this.edges = new LinkedList<Edge>();
                this.path = new LinkedList<Edge>();
                this.mapSize = mapSize;
                this.threshold = threshold;

                this.marginLeft = 20;
                this.marginRight = 100; // so that labels won't go outside map
                this.marginTop = 20;
                this.marginBottom = 20;

                this.backgroundImage = backgroundImage;

                createTransforms(panelSize);

                edgeStroke = new BasicStroke();
                pathStroke = new BasicStroke(5.0f);

                nodeLabelFont = new Font("SansSerif", Font.PLAIN, 12);
                pathLabelFont = new Font("SansSerif", Font.BOLD, 14);

                setPreferredSize(panelSize);
                setOpaque(true);
                setBackground(Color.white);

                addMouseListener(new ClickListener());
                addComponentListener(new ComponentAdapter() {
                                public void componentResized(ComponentEvent e) {
                                        updateSize();
                                }
                        });
        }

        public void setBackgroundImage(ImageIcon backgroundImage) {
                this.backgroundImage = backgroundImage;
                repaint();
        }

        private void updateSize() {
                createTransforms(getSize());
                repaint();
        }

        private void createTransforms(Dimension panelSize) {
                mapToScreenTransform = new AffineTransform();

                // note that these transforms are done "bottom to top"

                // move out to margin
                mapToScreenTransform.translate(marginLeft, marginTop);

                // scale to panel size and flip y-axis since graphics coordinates start at upper-left
                double xscale = (panelSize.getWidth() - marginLeft - marginRight) / mapSize.getWidth();
                double yscale = - (panelSize.getHeight() - marginTop - marginBottom) / mapSize.getHeight();
                mapToScreenTransform.scale(xscale, yscale);

                // move below x-axis
                mapToScreenTransform.translate(0.0, - mapSize.getHeight());

                try {
                        screenToMapTransform = mapToScreenTransform.createInverse();
                } catch (NoninvertibleTransformException ex) {
                        assert false : ex;
                }
        }

//         /**
//          *  Makes a point in screen space from map space coordinates.
//          *  Validates the 
//          */
//         private Point mkPoint(int x, int y) {
// /*
//                 if (x < 0 || y < 0 
//                     || x >= mapSize.width || y >= mapSize.height)
//                         throw new IllegalArgumentException(
//                                 "Bad coordinates: (" + x + "," + y + "), size: " + size);
// */
//                 return mapToScreen(new Point(x,y));
//         }

        private Point screenToMap(Point p) {
                Point r = new Point();
                screenToMapTransform.transform(p, r);
                return r;
        }

        private Point mapToScreen(Point p) {
                Point r = new Point();
                mapToScreenTransform.transform(p, r);
                return r;
        }

        public void addNode(String label, int x, int y) {
                nodes.add(new Node(label, x, y));
                repaint();
        }

        public void addNode(String label, Point p) {
                addNode(label, p.x, p.y);
        }

        public void addEdge(String label, int fromx, int fromy, int tox, int toy) {
                edges.add(new Edge(label, fromx, fromy, tox, toy));
                repaint();
        }

        public void addEdge(String label, Point from, Point to) {
                addEdge(label, from.x, from.y, to.x, to.y);
        }

        public void clearPath() {
                path.clear();
                repaint();
        }

        public void addPathEdge(String label, int fromx, int fromy, int tox, int toy) {
                path.add(new Edge(label, fromx, fromy, tox, toy));
                repaint();
        }

        public void addPathEdge(String label, Point from, Point to) {
                addPathEdge(label, from.x, from.y, to.x, to.y);
        }

        protected void paintComponent(Graphics gr) {
                super.paintComponent(gr);
                Graphics2D g = (Graphics2D)gr;

                g.setRenderingHint(RenderingHints.KEY_ANTIALIASING, 
                                   RenderingHints.VALUE_ANTIALIAS_ON);

                Color oldColor = g.getColor();

                int width = getWidth();
                int height = getHeight();
                if (isOpaque()) {
                        g.setColor(getBackground());
                        g.fillRect(0, 0, width, height);
                }
                if (backgroundImage != null) {
                        int imageWidth = width - marginLeft - marginRight;
                        int imageHeight = height - marginTop - marginBottom;
                        g.drawImage(backgroundImage.getImage(), marginLeft, marginBottom, 
                                    imageWidth, imageHeight, this);
                }

                AffineTransform saveTrans = g.getTransform();
                g.transform(mapToScreenTransform);

                g.setStroke(edgeStroke);

                // draw edges
                g.setColor(Color.gray);
                for (Edge e : edges) 
                        g.drawLine(e.from.x, e.from.y, e.to.x, e.to.y);

                // draw nodes
                for (Node n : nodes) {
                        g.setColor(Color.blue);
                        g.drawOval((int)(n.pos.x - threshold), (int)(n.pos.y - threshold),
                                   (int)(2*threshold), (int)(2*threshold));
                        g.setColor(Color.black);

                        g.fillOval(n.pos.x - NODE_RADIUS, n.pos.y - NODE_RADIUS, 
                                   2*NODE_RADIUS, 2*NODE_RADIUS);
                }

                // draw the path
                g.setColor(Color.red);
                g.setStroke(pathStroke);
                for (Edge e : path)
                        g.drawLine(e.from.x, e.from.y, e.to.x, e.to.y);

                g.setTransform(saveTrans);

                // draw node names
                g.setColor(Color.black);
                g.setFont(nodeLabelFont);
                for (Node n : nodes) {
                        // FIXME: calculate some nice offset
                        Point p = mapToScreen(new Point(n.pos.x + LABEL_X_OFF, n.pos.y + LABEL_Y_OFF));
                        g.drawString(n.label, p.x, p.y);
                }

                // draw path edge labels
                g.setColor(Color.blue);
                g.setFont(pathLabelFont);
                for (Edge e : path) {
                        // FIXME: not so pretty
                        Point p = mapToScreen(new Point((e.from.x + e.to.x) / 2, 
                                                        (e.from.y + e.to.y) / 2));
                        g.drawString(e.label, p.x, p.y);
                }

                g.setColor(oldColor);
        }

        public void addMapClickListener(MapClickListener l) {
                listenerList.add(MapClickListener.class, l);
        }

        public void removeMapClickListener(MapClickListener l) {
                listenerList.remove(MapClickListener.class, l);
        }

        /**
         *  @param p The point (in map coordinates) that was clicked.
         *  @param t The time (in milliseconds since the Unix epoch at which
         *  the click occured.
         */
        protected void fireMapClicked(Point p, long t) {
                MapClickEvent mapClickEvent = new MapClickEvent(this, p, t); 
                MapClickListener[] listeners = listenerList.getListeners(MapClickListener.class);
                for (int i = listeners.length - 1; i >= 0; i--)
                        listeners[i].mapClicked(mapClickEvent);
        }

        private class ClickListener extends MouseAdapter {
                public void mouseClicked(MouseEvent e) {
                        Point p = screenToMap(e.getPoint());
                        fireMapClicked(p, e.getWhen());
                }
        }

        private class Node {
                public String label;
                public Point pos;
                public Node(String label, int x, int y) {
                        this.label = label;
                        this.pos = new Point(x,y);
                }
        }

        private class Edge {
                public String label;
                public Point from;
                public Point to;
                public Edge(String label, int fromx, int fromy, int tox, int toy) {
                        this.label = label;
                        this.from = new Point(fromx, fromy);
                        this.to = new Point(tox, toy);
                }
        }

}
