package se.chalmers.cs.gf.tramdemo;

import java.awt.Dimension;
import java.awt.Point;
import java.net.URL;
import java.util.*;
import java.util.regex.*;
import javax.swing.*;
import javax.swing.event.EventListenerList;

import se.chalmers.cs.gf.graph.*;

public class TramMap extends JFrame {

        protected EventListenerList listenerList = new EventListenerList();

        private TransportNetwork network;

        private CoordMap<String> coordMap;

        private MapPanel mapPanel;

        public TramMap(double clickThreshold, 
                       TransportNetwork network,
                       String backgroundImageResource) {
                super("Map");

                this.network = network;

                /// build coordinate map
                coordMap = network.getCoordMap(clickThreshold);

                // load background image
                ImageIcon backgroundImage = null;
                if (backgroundImageResource != null) {
                        URL bgURL = TramMap.class.getResource(backgroundImageResource);
                        if (bgURL != null) {
                                backgroundImage = new ImageIcon(bgURL);
                        } else {
                                System.err.println("Background image " + backgroundImageResource + " not found");
                        }
                }

                // calculate width and height of network
                int width = 0, height = 0;
                for (TransportNetwork.Node n : network.nodes()) {
                        Point p = n.getValue().getPos();
                        width = Math.max(p.x, width);
                        height = Math.max(p.y, height);
                }

                // build map panel
                Dimension mapSize = new Dimension(width, height);
                Dimension panelSize = new Dimension(600,400);
                this.mapPanel = new MapPanel(mapSize, panelSize, clickThreshold, backgroundImage);
                for (TransportNetwork.Node n : network.nodes()) {
                        mapPanel.addNode(n.getValue().getName(), n.getValue().getPos());
                        for (TransportNetwork.Edge e : n.adjacent()) {
                                mapPanel.addEdge(e.getValue(),
                                                 e.getFrom().getValue().getPos(), 
                                                 e.getTo().getValue().getPos());
                        }
                }
                mapPanel.addMapClickListener(new ClickListener());

                setContentPane(mapPanel);
        }

        public void addMapClickListener(MapClickListener l) {
                listenerList.add(MapClickListener.class, l);
        }

        public void removeMapClickListener(MapClickListener l) {
                listenerList.remove(MapClickListener.class, l);
        }

        /**
         *  @param p The point (in map coordinates) that was clicked.
         *  @param regions The regions which the click was in
         *  @param t The time (in milliseconds since the Unix epoch at which
         *  the click occured.
         */
        protected void fireMapClicked(Point p, Collection<String> regions, long t) {
                MapClickEvent mapClickEvent = new MapClickEvent(this, p, regions, t); 
                MapClickListener[] listeners = listenerList.getListeners(MapClickListener.class);
                for (int i = listeners.length - 1; i >= 0; i--)
                        listeners[i].mapClicked(mapClickEvent);
        }

        private class ClickListener implements MapClickListener {
                public void mapClicked(MapClickEvent e) {
                        Collection<String> regions = coordMap.get(e.getPoint());
                        fireMapClicked(e.getPoint(), regions, e.getTime());
                }
        }

        /**
         *  Perform a number of drawing instructions.
         *  Maybe called from any thread.
         *  Example input:  "clear; drawEdge (1, [Gronsakstorget, Hagakyrkan, Jarntorget]);"
         */
        public void doDrawInstrs(final String instrs) {    
                System.err.println("TramMap: got instructions: " + instrs);
            
                SwingUtilities.invokeLater(new Runnable() {
                                public void run() {
                                        doDrawInstrs_(instrs);
                                }
                        });
        }

        private Pattern drawEdgePatt = Pattern.compile("drawEdge\\s*\\(\\s*([^,]+)\\s*,\\s*\\[\\s*([^\\]]*)\\s*\\]\\s*\\)");

        private void doDrawInstrs_(String instrs) {
                String[] is = instrs.split("\\s*;\\s*");
                for (String i : is) {
                        if (i.length() == 0)
                                continue;

                        if (i.equals("clear")) {
                                clearPath_();
                                continue;
                        }

                        Matcher m = drawEdgePatt.matcher(i);
                        if (m.find()) {
                                String label = m.group(1);
                                String stopList = m.group(2);
                                String[] stops = stopList.split("\\s*,\\s*");
                                for (int x = 1; x < stops.length; x++)
                                        showEdge_(label, stops[x-1], stops[x]);
                                continue;
                        }

                        System.err.println("Unknown instruction " + i);
                }
        }

        /**
         * May be called from any thread.
         */
        public void showEdge(final String label, final String from, final String to) {
                SwingUtilities.invokeLater(new Runnable() {
                                public void run() {
                                        showEdge_(label, from, to);
                                }
                        });
        }

        private Point getNodePos(String name) {
                return network.getNode(name).getValue().getPos();
        }

        private void showEdge_(String label, String from, String to) {
                Point f = getNodePos(from);
                if (f == null) {
                        System.err.println("No such stop: " + from);
                        return;
                }                        
                Point t = getNodePos(to);
                if (t == null) {
                        System.err.println("No such stop: " + to);
                        return;
                }
                mapPanel.addPathEdge(label, f, t);
        }

        /**
         * May be called from any thread.
         */
        public void clearPath() {
                SwingUtilities.invokeLater(new Runnable() {
                                public void run() {
                                        clearPath_();
                                }
                        });
        }

        private void clearPath_() {
                mapPanel.clearPath();
        }

        public static TramMap showMap(double clickThreshold, 
                                      TransportNetwork network, 
                                      String backgroundImageResource) {
                TramMap map = new TramMap(clickThreshold, network, 
                                          backgroundImageResource);
                map.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
                map.pack();
                map.setVisible(true);
                return map;
        }

}
